/*
Manage platform-independent loading of shared libraries.
*/

#pragma once
#include <ConsumerImplHelper/Platform.h>
#include <string>
#if defined CIH_WIN_BUILD
#   include <windows.h>
#elif defined CIH_LINUX_BUILD
#   include <dlfcn.h>
#endif


namespace GenTLConsumerImplHelper
{
    namespace Internal
    {

        class CDynamicLib
        {
        public:
            CDynamicLib(void);

            ~CDynamicLib(void);

        public:

            ///////////////////////////////////////////////////////////////////////
            bool Load(const char* pName);


            ///////////////////////////////////////////////////////////////////////
            bool Unload();


            ///////////////////////////////////////////////////////////////////////
            void* GetFunctionPointer(const char* pszName);


            ///////////////////////////////////////////////////////////////////////
            bool IsValid() const
            {
                return m_hDLL != NULL;
            }


            ///////////////////////////////////////////////////////////////////////
            const char* GetLastErrorMsg() const
            {
                return m_error.c_str();
            }


#if defined CIH_WIN_BUILD
        private:
            void SetErrorMsg(DWORD errorCode);
#endif

        private:
#if defined CIH_WIN_BUILD
            HMODULE m_hDLL;
#elif defined CIH_LINUX_BUILD
            void* m_hDLL;
#endif
            std::string m_error;
        };


        inline CDynamicLib::CDynamicLib( void ) : m_hDLL( NULL )
        {
        }

        inline CDynamicLib::~CDynamicLib( void )
        {
            Unload();
        }

        inline bool CDynamicLib::Load( const char* pName )
        {
#if defined CIH_WIN_BUILD
            HMODULE h = ::LoadLibraryExA(pName, NULL, LOAD_WITH_ALTERED_SEARCH_PATH);

            if (h == NULL)
                SetErrorMsg(::GetLastError());
            else
                m_error = "";

#elif defined CIH_LINUX_BUILD

            void* h = dlopen(pName, RTLD_NOW | RTLD_LOCAL);

            const char* error = dlerror();
            m_error = error ? error : "";
#endif

            if (h == NULL)
            {
                // error loading DLL
                return false;
            }

            // after we successfully loaded the DLL
            // unload prev lib and set the member
            Unload();

            m_hDLL = h;


            return true;
        }

        inline bool CDynamicLib::Unload()
        {
            if (IsValid())
            {
                // m_error only has to be set when returning false on error.
#if defined CIH_WIN_BUILD
                ::FreeLibrary(m_hDLL);
#elif defined CIH_LINUX_BUILD
                dlclose(m_hDLL);
#endif
                m_hDLL = NULL;
                return true;
            }
            else
            {
                return false;
            }
        }

        inline void* CDynamicLib::GetFunctionPointer( const char* pszName )
        {
            if (!IsValid())
            {
                return NULL;
            }

#if defined CIH_WIN_BUILD
            void* pFunc(::GetProcAddress(m_hDLL, pszName));

            if (pFunc == NULL)
                SetErrorMsg(::GetLastError());
            else
                m_error = "";

#elif defined CIH_LINUX_BUILD
            dlerror();  // Clear dlerror() to distinguish between "symbol not found" and an error after dlsym()

            void* pFunc;
            *(void**)(&pFunc) = dlsym(m_hDLL, pszName);

            const char* error = dlerror();
            m_error = error ? error : "";
#endif

            return pFunc;
        }

#if defined CIH_WIN_BUILD
        inline void CDynamicLib::SetErrorMsg( DWORD errorCode )
        {
            LPSTR pMsgBuf = NULL;
            if (::FormatMessageA(
                FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
                NULL,
                errorCode,
                MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                (LPSTR)&pMsgBuf,
                0, NULL))
            {
                m_error = pMsgBuf;
                LocalFree(pMsgBuf);
            }
            else
            {
                m_error = "FormatMessage failed for error code 0x";
                char buffer[sizeof(long) * 2 + 1];
                _ltoa_s( errorCode, buffer, 16);
                m_error += buffer;
            }
        }
#endif

    }
}
