#pragma once

#include "Platform.h"

#ifdef __cplusplus
extern "C" {
#endif

#if defined(CIH_WIN_BUILD)
#  include <Windows.h>
#  include <errno.h>

#  ifndef ETIMEDOUT
#     define ETIMEDOUT       110
#  endif

  typedef HANDLE WaitObject_t;
#  define INVALID_WAIT_OBJECT INVALID_HANDLE_VALUE

#elif defined (CIH_LINUX_BUILD)
#  include <unistd.h>
#  include <errno.h>
#  include <stdlib.h>
#  include <string.h>
#  include <fcntl.h>
#  include <sys/time.h>
#  include <poll.h>
#  include <limits.h>

#  undef INFINITE
#  define INFINITE            0xFFFFFFFF

  typedef unsigned long TimePoint_t;
  typedef signed long TimePeriod_t;

  typedef struct _WaitObjectPosix_t
  {
     int m_fd[2];
  } WaitObjectPosix_t;

  typedef WaitObjectPosix_t* WaitObject_t;
#  define INVALID_WAIT_OBJECT (0l)
#else 
#   error Unsupported plattform!
#endif

  /**
      \brief Creates a wait object
      This wait objects is NOT automatically reset after beeing signalled.
      On windows this is equivalent to calling CreateEvent(NULL, TRUE, initiallySignalled, NULL)

      \return a wait object handle or NULL on error
  */
  static inline WaitObject_t CreateWaitObject(int initiallySignalled);

  /**
      \brief Closes the given WaitObject
  */
  static inline void CloseWaitObject(WaitObject_t pWaitObject);

  /**
      \brief Duplicates a wait object
      The signalled state is duplicated also
      \return a wait object handle or NULL on error
  */
  static inline WaitObject_t DuplicateWaitObject(WaitObject_t pSrcWaitObject);

  /**
      \brief Signals the given wait object
      \return 0 on success, -1 on error.
  */
  static inline int SignalWaitObject(WaitObject_t pWaitObject);

  /**
      \brief Resets the given wait object
      \return 0 on success, -1 on error.
  */
  static inline int ResetWaitObject(WaitObject_t pWaitObject);

  /**
      \brief Checks if the given wait object is signalled
      \return 1 if the wait object is signalled, 0 otherwise.
  */
  static inline int WaitObjectIsSignalled(WaitObject_t pWaitObject);

  /**
      \brief Wait for the given wait object
      \param pWaitObject The wait object
      \param timeout The timeout in milliseconds, INFINITE waits forever
      \param bAlertable set this to 1 if the wait should be interruptable by a signal
      \return 0 if the wait object got signalled
              EINTR if the wait was interrupted
              ETIMEDOUT if the wait timed out
              -1 on error
  */
  static inline int WaitExWaitObjectWithTimeout(WaitObject_t pWaitObject, unsigned int timeout, int bAlertable);

  /**
      \brief Wait for the given wait object
      \param pWaitObject The wait object
      \param bAlertable set this to 1 if the wait should be interruptable by a signal
      \return 0 if the wait object got signalled
              EINTR if the wait was interrupted
              -1 on error
  */
  static inline int WaitExWaitObject(WaitObject_t pWaitObject, int bAlertable);

  /**
      \brief Wait for multiple wait objects
      \param count The number of objects in objects
      \param objects Array of wait objects
      \param timeout The timeout in milliseconds, INFINITE waits forever
      \param bAlertable set this to 1 if the wait should be interruptable by a signal
      \param bWaitAll Set this to 1 if you want to wait for all objects. This does not guarantee that all objects where signalled at the same time.
                      If this is 0, the wait ends when the first object is signalled and pIndex is set to the corresponding index
      \param pIndex The index of the wait object which ended the wait. This is only set if bWaitAll is 0.
      \return 0 if the wait object got signalled
              EINTR if the wait was interrupted
              ETIMEDOUT if the wait timed out
              -1 on error
  */
  static inline int WaitExWaitMultipleObjectsWithTimeout(unsigned count, WaitObject_t* objects, unsigned int timeout, int bAlertable, int bWaitAll, unsigned *pIndex);

  static inline int WaitObjectIsSignalled(WaitObject_t pWaitObject) {
      if(WaitExWaitObjectWithTimeout( pWaitObject, 0, 0) == 0)
      {
          return 1;
      }

      return 0;
  }




#if defined(CIH_WIN_BUILD)
    // Note: In contrast to the WinAPI functions, these functions return errno.h constants. So a return value of 0 means no error.

    //simple functions to cast the type
    static inline HANDLE WaitObjectToHandle(WaitObject_t waitObject) {
        return (HANDLE)waitObject;
    }

    static inline WaitObject_t WaitObjectFromHandle(HANDLE handle) {
        return (WaitObject_t)handle;
    }


    static WaitObject_t CreateWaitObject(int initiallySignalled)
    {
        return ::CreateEvent(NULL, TRUE, initiallySignalled, NULL);
    }

    static void CloseWaitObject(WaitObject_t waitObject)
    {
        ::CloseHandle(waitObject);
    }

    static WaitObject_t DuplicateWaitObject(WaitObject_t pSrcWaitObject)
    {
        WaitObject_t dup = NULL;
        const HANDLE hProc = ::GetCurrentProcess();
        if (!::DuplicateHandle(hProc, pSrcWaitObject, hProc, &dup, 0, false, DUPLICATE_SAME_ACCESS)) {
            dup = NULL;
        }

        return dup;
    }

    static int SignalWaitObject(WaitObject_t waitObject) {
        if(::SetEvent(waitObject) == 0)
        {
            return ENOMEM;
        }
        return 0;
    }

    static int ResetWaitObject(WaitObject_t waitObject) {
        if(::ResetEvent(waitObject) == 0)
        {
            return ENOMEM;
        }
        return 0;
    }

    static int WaitExWaitObjectWithTimeout(WaitObject_t pWaitObject, unsigned int timeout, int bAlertable) {
        DWORD res = ::WaitForSingleObjectEx(pWaitObject, timeout, bAlertable);
        switch (res)
        {
        case WAIT_OBJECT_0:
            return 0;
        case WAIT_TIMEOUT:
            return ETIMEDOUT;
        case WAIT_IO_COMPLETION:
            return EINTR;
        default:
            return ENOMEM;
        }
    }

    static int WaitExWaitObject(WaitObject_t pWaitObject, int bAlertable)
    {
        return WaitExWaitObjectWithTimeout(pWaitObject, INFINITE, bAlertable);
    }

    // ----------------------------------------------------------------------
    static int WaitExWaitMultipleObjectsWithTimeout(unsigned count, WaitObject_t* objects, unsigned int timeout, int bAlertable, int bWaitAll, unsigned *pIndex)
    {
        DWORD res;
        res = ::WaitForMultipleObjectsEx(count, objects, bWaitAll, timeout, bAlertable);

        // Check for errors
        switch (res)
        {
        case WAIT_TIMEOUT:
            return ETIMEDOUT;
        case WAIT_IO_COMPLETION:
            return EINTR;
        case WAIT_FAILED:
            return ENOMEM;
        }

        // Check which object was signaled
        DWORD idx0 = res - WAIT_OBJECT_0;
        if (idx0 < count)
        {
            if (pIndex != NULL)
            {
                *pIndex = idx0;
            }
            return 0;
        }

        // Check which object was abandoned.
        DWORD idx1 = res - WAIT_ABANDONED_0;
        if (idx1 < count)
        {
            if (pIndex != NULL)
            {
                *pIndex = idx1;
            }
            return 0;
        }

        //don't know what happened
        return -1;
    }



#elif defined (CIH_LINUX_BUILD)

static __inline uint32_t mtime(void)
{
    uint32_t time_ms = UINT_MAX;

    struct timespec total; // intentionally not initialized

    clock_gettime(CLOCK_MONOTONIC, &total);

    time_ms = (uint32_t)(total.tv_sec * 1000 + total.tv_nsec / 1000000);

    return time_ms;
}


static inline int WaitExWaitMultipleObjects(unsigned count, WaitObject_t* objects, int bAlertable, int bWaitAll, unsigned *pIndex);

// ----------------------------------------------------------------------
static WaitObject_t CreateWaitObject(int initiallySignalled)
{
    int rc = -1;
    int rcErrno = 0;
    WaitObject_t pWaitObject = NULL;

    // create empty wait object
    pWaitObject = (WaitObject_t)malloc(sizeof(WaitObjectPosix_t));
    if (NULL == pWaitObject)
    {
        rc = -1;
        rcErrno = errno;
        goto Exit;
    }
    pWaitObject->m_fd[0] = pWaitObject->m_fd[1] = -1;

    // create wait object pipe and set pipe parameter
    rc = pipe(pWaitObject->m_fd);
    if (0 != rc)
    {
        rcErrno = errno;
        goto Exit;
    }
    rc =  fcntl(pWaitObject->m_fd[0], F_SETFL, fcntl(pWaitObject->m_fd[0],F_GETFL, 0) | O_NONBLOCK);
    if (0 != rc)
    {
        rcErrno = errno;
        goto Exit;
    }

    // signal wait object if requested
    if (initiallySignalled)
    {
        rc = SignalWaitObject(pWaitObject);
        if (0 != rc)
        {
            rcErrno = errno;
        }
    }

Exit:
    if (0 != rc)
    {
        CloseWaitObject(pWaitObject);
        pWaitObject = NULL;
        errno = rcErrno;
    }

    return pWaitObject;
}

// ----------------------------------------------------------------------
// Create a wait object with from a file descriptor which can be used by select.
// The wait object is signaled from outside by the user by writing to the fd.
// A call to SignalWaitObject() will fail, as we have no writing pipe end.
// ResetWaitObject(), which reads, should not be called, as the data belongs to the user,
// but to be robust we set the file status flags of the duplicate to O_NONBLOCK.
static inline WaitObject_t CreateWaitObjectFromFd(int fd)
{
    int rc = -1;
    int rcErrno = 0;
    WaitObject_t pWaitObject;

    // create empty wait object
    pWaitObject = (WaitObject_t)malloc(sizeof(WaitObjectPosix_t));
    if (NULL == pWaitObject)
    {
        rc = -1;
        rcErrno = errno;
        goto Exit;
    }

    // set writing fd to invalid (we can not signal this fd from inside)
    pWaitObject->m_fd[1] = -1;

    // set polling fd to a duplicated requested one.
    pWaitObject->m_fd[0] = dup(fd);
    if (pWaitObject->m_fd[0] < 0)
    {
        rc = -1;
        rcErrno = errno;
        goto Exit;
    }

    // do not block when ResetWaitObject() is called
    rc = fcntl(pWaitObject->m_fd[0], F_SETFL, fcntl(pWaitObject->m_fd[0],F_GETFL, 0) | O_NONBLOCK);
    if (0 != rc)
    {
        rcErrno = errno;
        goto Exit;
    }

Exit:
    if (0 != rc)
    {
        CloseWaitObject(pWaitObject);
        pWaitObject = NULL;
        errno = rcErrno;
    }

    return pWaitObject;
}

// ----------------------------------------------------------------------
static void CloseWaitObject(WaitObject_t pWaitObject)
{
    if (pWaitObject != NULL)
    {
        if (pWaitObject->m_fd[0] >= 0) close(pWaitObject->m_fd[0]);
        if (pWaitObject->m_fd[1] >= 0) close(pWaitObject->m_fd[1]);
        free(pWaitObject);
    }
}

// ----------------------------------------------------------------------
WaitObject_t DuplicateWaitObject(WaitObject_t pSrcWaitObject)
{
    int rc = -1;
    int rcErrno = 0;
    WaitObject_t pWaitObject = NULL;

    if (pSrcWaitObject == NULL)
    {
        rc = EINVAL;
        goto Exit;
    }

    // create empty wait object
    pWaitObject = (WaitObject_t)malloc(sizeof(WaitObjectPosix_t));
    if (NULL == pWaitObject)
    {
        rc = -1;
        rcErrno = errno;
        goto Exit;
    }
    pWaitObject->m_fd[0] = pWaitObject->m_fd[1] = -1;

    // duplicate file descriptors (0)
    if (pSrcWaitObject->m_fd[0] >= 0)
    {
        pWaitObject->m_fd[0] = dup(pSrcWaitObject->m_fd[0]);
        if (pWaitObject->m_fd[0] < 0)
        {
            rc = -1; rcErrno = errno;
            goto Exit;
        }
    }

    // duplicate file descriptors (1)
    if (pSrcWaitObject->m_fd[1] >= 0)
    {
        pWaitObject->m_fd[1] = dup(pSrcWaitObject->m_fd[1]);
        if (pWaitObject->m_fd[1] < 0)
        {
            rc = -1; rcErrno = errno;
            goto Exit;
        }
    }

    rc = 0;

Exit:
    if (0 != rc)
    {
        CloseWaitObject(pWaitObject);
        pWaitObject = NULL;
        errno = rcErrno;
    }

    return pWaitObject;
}

// ----------------------------------------------------------------------
static int SignalWaitObject(WaitObject_t pWaitObject)
{
    int rc = -1;
    int rcErrno = 0;
    static const char data = 0;

    if (pWaitObject == NULL)
    {
        errno = EINVAL;
        return -1;
    }

    //check if the object is already signaled
    struct pollfd fds;
    fds.fd = pWaitObject->m_fd[0];
    fds.events = POLLIN;
    fds.revents = 0;

    do
    {
        rc = poll(&fds, 1, 0);
    }
    while (rc < 0 && errno == EINTR);

    if(rc == 0) //timeout, this means not signaled
    {
        do
        {
            rc = write(pWaitObject->m_fd[1], &data, 1);
        }
        while (rc == -1L && errno == EINTR);

        if(rc == 1) //write successful
        {
            rc = 0;
        }
        else //write failed
        {
            rc = -1;
            rcErrno = errno;
        }
    }
    else if (rc < 0) //polling failed
    {
        rc = -1;
        rcErrno = errno;
    }
    else //the object is already signalled
    {
        rc = 0;
    }

    if (0 != rc)
    {
        errno = rcErrno;
    }

    return rc;
}

// ----------------------------------------------------------------------
static int ResetWaitObject(WaitObject_t pWaitObject)
{
    int rc = -1;
    int rcErrno = 0;
    char data;

    if (pWaitObject == NULL)
    {
        rc = -1;
        rcErrno = EINVAL;
        goto Exit;
    }

    do {
        rc = read(pWaitObject->m_fd[0], &data, 1);
    }
    while (rc == -1L && errno == EINTR);

    if (0 > rc && errno != EAGAIN)
    {
        rc = -1; rcErrno = errno;
        goto Exit;
    }

    // reset was successful
    rc = 0;

Exit:
    if (0 != rc)
    {
        errno = rcErrno;
    }

    return rc;
}

// ----------------------------------------------------------------------
static int WaitExWaitObjectWithTimeout(WaitObject_t pWaitObject, unsigned int timeout, int bAlertable)
{
    // Check if timeout is not INFINITE
    if (timeout != INFINITE)
    {
        struct pollfd fds;
        const TimePoint_t end = mtime() + timeout;


        fds.fd = pWaitObject->m_fd[0];
        fds.events = POLLIN;
        fds.revents = 0;

        //we need to run at least once, even if timeout is 0
        do
        {
            int nrdy;

            TimePoint_t cur = mtime();
            TimePeriod_t remaining_ms = ((cur > end) ? 0 : (end - cur));

            nrdy = poll(&fds, 1, remaining_ms );
            if (nrdy < 0) //polling failed
            {
                if (errno == EINTR) // poll was interrupted
                {
                    if (!bAlertable)
                    {
                        continue;
                    }
                    return EINTR;
                }

                // error
                return -1;
            }
            else if(nrdy == 0) // timeout
            {
                errno = ETIME;
                return ETIMEDOUT;
            }
            else
            {
                //fd was signalled
                errno = 0;
                return 0;
            }
        }
        while (mtime() < end);

        //id we end up here, the time was up
        errno = ETIME;
        return ETIMEDOUT;
    }
    else
    {
        // Call implementation without timeout
        return WaitExWaitObject(pWaitObject, bAlertable);
    }
}

// ----------------------------------------------------------------------
static int WaitExWaitObject(WaitObject_t pWaitObject, int bAlertable)
{
    struct pollfd fds;
    int nrdy;

    fds.fd = pWaitObject->m_fd[0];
    fds.events = POLLIN;
    fds.revents = 0;

    do
    {
        nrdy = poll(&fds, 1, -1);
        if (nrdy < 0)
        {
            if (errno == EINTR) // poll was interrupted by a signal
            {
                if (!bAlertable)
                {
                    continue;
                }
                return EINTR; // waitex_alerted
            }

            // error
            return -1;
        }
    }
    while (nrdy < 0);

    //per definition, this can't happen, because we poll with infinite timeout
    if (nrdy == 0)
    {
        errno = ETIME;
        return -1;
    }

    return 0; // waitex_signaled
}

#  ifdef __GNUC__
// suppress warnings about variable length arrays used by the WaitObject functions
#  pragma GCC diagnostic push
#  pragma GCC diagnostic ignored "-Wvla"
#  endif

// ----------------------------------------------------------------------
// in case of bWaitAll == true, the implementation returns true when during the timeout period
// every object has been signaled at least once.
// The implementation doesn't check if there was a moment when all objects have been signaled
// simultaneously
static int WaitExWaitMultipleObjectsWithTimeout(unsigned count, WaitObject_t* objects, unsigned int timeout, int bAlertable, int bWaitAll, unsigned *pIndex)
{
    // Check if timeout is not INFINITE
    if (timeout != INFINITE)
    {
        unsigned int i;
        const TimePoint_t end = mtime() + timeout;
        int nOutstanding = count;

        //this should work in C99
        struct pollfd fds[count];
        for(i = 0; i < count; i++)
        {
            fds[i].fd = objects[i]->m_fd[0];
            fds[i].events = POLLIN;
            fds[i].revents = 0;
        }

        do
        {
            TimePoint_t cur = mtime();
            TimePeriod_t remaining_ms = ((cur > end) ? 0 : (end - cur));

            const int nrdy = poll(fds, count, remaining_ms );
            if (nrdy < 0) // poll failed
            {
                if (errno == EINTR) // poll was interrupted by a signal
                { 
                    if (!bAlertable)
                    {
                        continue;
                    }
                    return EINTR;
                }

                return -1;
            }
            else if(nrdy == 0) // poll timed out
            { 
                return ETIMEDOUT;
            }

            for(i = 0; i < count; i++)
            {
                if ( fds[i].revents )
                {
                    //early out, if we wait only for one
                    if ( !bWaitAll )
                    {
                        if ( pIndex )
                        {
                            *pIndex = i;
                        }

                        return 0;
                    }

                    //ignore this fd in the next poll
                    fds[i].fd = -1;
                    fds[i].revents = 0;
                    nOutstanding--;
                }
            }
        } while ( end > mtime() && (nOutstanding > 0) );

        if(nOutstanding == 0)
        {
            return 0; // waitex_signaled
        }
        else
        {
            return ETIMEDOUT;
        }
    }
    else
    {
        // Call implementation without timeout
        return WaitExWaitMultipleObjects(count, objects, bAlertable, bWaitAll, pIndex);
    }
}

// ----------------------------------------------------------------------
static int WaitExWaitMultipleObjects(unsigned count, WaitObject_t* objects, int bAlertable, int bWaitAll, unsigned *pIndex)
{
    unsigned int i;
    int nrdy;
    int nOutstanding = count;

    //this should work in C99
    struct pollfd fds[count];
    for(i = 0; i < count; i++)
    {
        fds[i].fd = objects[i]->m_fd[0];
        fds[i].events = POLLIN;
        fds[i].revents = 0;
    }

    do
    {
        nrdy = poll(fds, count, -1);
        if (nrdy < 0)
        {
            if (errno == EINTR)
            { 
                // Signal
                if (!bAlertable)
                {
                    continue;
                }

                return EINTR;
            }

            return -1;
        }
        else if (nrdy == 0) //per definition this must not happen, because we called poll without timeout
        {
            errno = ETIME;
            return -1;
        }

        for(i = 0; i < count; i++)
        {
            if ( fds[i].revents )
            {
                //early out, if we wait only for one
                if ( !bWaitAll )
                {
                    if ( pIndex )
                    {
                        *pIndex = i;
                    }

                    return 0;
                }

                //ignore this fd in the next poll
                fds[i].fd = -1;
                fds[i].revents = 0;
                nOutstanding--;
            }
        }
    } while ( nOutstanding > 0 );

    return 0; // all objects were signalled
}

#ifdef __GNUC__
#pragma GCC diagnostic pop
#endif

#endif

#ifdef __cplusplus
}
#endif


